from __future__ import annotations

import platform
import shutil
import subprocess
from pathlib import Path

import pytest
import rasterio

import pywbt

is_linux = platform.system() == "Linux"

# To avoid redownloading and hitting the WBT server multiple times for testing
# on different platforms and Python versions, especially on CI, the binaries are
# stored in `tests/wbt_zip` directory.


def assert_results(output_tiff: Path, expected: float) -> None:
    with rasterio.open(output_tiff) as src:
        assert src.width == 233
        assert src.height == 303
        data = src.read(1)
        data[data == src.nodata] = 0
        assert abs(data.mean() - expected) < 1e-4


@pytest.fixture
def wbt_args() -> dict[str, list[str]]:
    return {
        "BreachDepressions": ["-i=dem.tif", "--fill_pits", "-o=dem_corr.tif"],
        "D8FlowAccumulation": ["-i=dem_corr.tif", "-o=d8accum.tif"],
        "ExtractStreams": ["--flow_accum=d8accum.tif", "--threshold=600.0", "-o=streams.tif"],
    }


def test_leastcost(temp_dir: str, wbt_zipfile: str) -> None:
    wbt_args = {
        "BreachDepressionsLeastCost": ["-i=dem.tif", "--fill", "--dist=100", "-o=dem_corr.tif"],
        "D8FlowAccumulation": ["-i=dem_corr.tif", "-o=d8accum.tif"],
        "ExtractStreams": ["--flow_accum=d8accum.tif", "--threshold=600.0", "-o=streams.tif"],
    }
    shutil.copy("tests/dem.tif", temp_dir)
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        save_dir=temp_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path("tests/wbt_zip") / wbt_zipfile,
    )
    assert_results(Path(temp_dir) / "streams.tif", 0.1149)


def test_same_src_dir(temp_dir: str, wbt_zipfile: str, wbt_args: dict[str, list[str]]) -> None:
    shutil.copy("tests/dem.tif", temp_dir)
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        ["streams.tif"],
        save_dir=temp_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path("tests/wbt_zip") / wbt_zipfile,
    )
    assert_results(Path(temp_dir) / "streams.tif", 0.1243)


def test_no_save(temp_dir: str, wbt_zipfile: str, wbt_args: dict[str, list[str]]) -> None:
    shutil.copy("tests/dem.tif", temp_dir)
    results_dir = Path(temp_dir) / "results"
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        save_dir=results_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path("tests/wbt_zip") / wbt_zipfile,
    )
    assert_results(results_dir / "streams.tif", 0.1243)


def test_with_save(temp_dir: str, wbt_zipfile: str, wbt_args: dict[str, list[str]]) -> None:
    shutil.copy("tests/dem.tif", temp_dir)
    results_dir = Path(temp_dir) / "results"
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        ["streams.tif"],
        save_dir=results_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path("tests/wbt_zip") / wbt_zipfile,
    )
    assert_results(results_dir / "streams.tif", 0.1243)


def test_wbt_error(temp_dir: str, wbt_zipfile: str, wbt_args: dict[str, list[str]]) -> None:
    wbt_args["ExtractStreams"][0] = "--flow_accum="
    shutil.copy("tests/dem.tif", temp_dir)
    with pytest.raises(RuntimeError):
        pywbt.whitebox_tools(
            temp_dir,
            wbt_args,
            save_dir=Path(temp_dir) / "results",
            wbt_root=Path(temp_dir) / "WBT",
            zip_path=Path("tests/wbt_zip") / wbt_zipfile,
        )


def test_wbt_wrong_output(temp_dir: str, wbt_zipfile: str, wbt_args: dict[str, list[str]]) -> None:
    shutil.copy("tests/dem.tif", temp_dir)
    with pytest.raises(FileNotFoundError):
        pywbt.whitebox_tools(
            temp_dir,
            wbt_args,
            ["no_streams.tif"],
            save_dir=Path(temp_dir) / "results",
            wbt_root=Path(temp_dir) / "WBT",
            zip_path=Path("tests/wbt_zip") / wbt_zipfile,
        )


def test_wrong_platform(
    temp_dir: str, wrong_wbt_zipfile: str, wbt_args: dict[str, list[str]]
) -> None:
    shutil.copy("tests/dem.tif", temp_dir)
    shutil.copy(Path("tests/wbt_zip") / wrong_wbt_zipfile, Path(temp_dir) / wrong_wbt_zipfile)
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        ["streams.tif"],
        save_dir=temp_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path(temp_dir) / wrong_wbt_zipfile,
    )
    assert_results(Path(temp_dir) / "streams.tif", 0.1243)


def test_tools() -> None:
    tools = pywbt.list_tools()
    assert "BreachDepressions" in tools
    for t in tools:
        assert isinstance(pywbt.tool_parameters(t), list)


def test_tif_to_gdf(temp_dir: str, wbt_zipfile: str) -> None:
    wbt_args = {
        "BreachDepressions": ["-i=dem.tif", "--fill_pits", "-o=dem_corr.tif"],
        "D8Pointer": ["-i=dem_corr.tif", "-o=fdir.tif"],
        "Basins": ["--d8_pntr=fdir.tif", "-o=basins.tif"],
    }
    shutil.copy("tests/dem.tif", temp_dir)
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        ["basins.tif"],
        save_dir=temp_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path(temp_dir) / wbt_zipfile,
    )
    basin_geo = pywbt.dem_utils.tif_to_gdf(Path(temp_dir) / "basins.tif", "int32", "basin")
    assert basin_geo.area.idxmax() == 175


def test_shp_out(temp_dir: str, wbt_zipfile: str) -> None:
    wbt_args = {
        "BreachDepressions": ["-i=dem.tif", "--fill_pits", "-o=dem_corr.tif"],
        "D8Pointer": ["-i=dem_corr.tif", "-o=fdir.tif"],
        "D8FlowAccumulation": ["-i=fdir.tif", "--pntr", "-o=d8accum.tif"],
        "ExtractStreams": ["--flow_accum=d8accum.tif", "--threshold=600.0", "-o=streams.tif"],
        "RasterToVectorLines": ["-i=streams.tif", "-o=streams.shp"],
    }
    shutil.copy("tests/dem.tif", temp_dir)
    pywbt.whitebox_tools(
        temp_dir,
        wbt_args,
        ["streams.shp"],
        save_dir=temp_dir,
        wbt_root=Path(temp_dir) / "WBT",
        zip_path=Path(temp_dir) / wbt_zipfile,
    )
    assert Path(temp_dir, "streams.shp").stat().st_size == 179412


def test_wrong_res() -> None:
    bbox = (-95.20, 29.70, -95.201, 29.701)
    with pytest.raises(ValueError, match="Resolution must be one of 10, 30, or 60 meters."):
        pywbt.dem_utils.get_3dep(bbox, "3dep.tif", resolution=29, to_5070=True)


@pytest.mark.xfail(is_linux, reason="pyproj seem to have issues on Linux.")
def test_dem_utils(temp_dir: str) -> None:
    bbox = (-95.201, 29.70, -95.20, 29.701)
    fname_3dep = Path(temp_dir) / "3dep.tif"
    pywbt.dem_utils.get_3dep(bbox, fname_3dep, resolution=30, to_5070=True)
    d3 = pywbt.dem_utils.tif_to_da(fname_3dep)
    fname_nasadem = Path(temp_dir) / "nasadem.tif"
    pywbt.dem_utils.get_nasadem(bbox, fname_nasadem, to_utm=True)
    dn = pywbt.dem_utils.tif_to_da(fname_nasadem, "int16", "elevation", "Elevation", -32768)
    assert d3.shape == (4, 4)
    assert dn.shape == (5, 5)
    assert d3.mean().item() == pytest.approx(dn.mean().item(), rel=1.3)


@pytest.fixture
def valid_toml_file() -> Path:
    return Path("tests/config.toml")


@pytest.fixture
def invalid_toml_file_not_dict(tmp_path: Path) -> Path:
    content = """
    src_dir = "data/input"
    arg_dict = "not_a_dict"
    """
    toml_file = tmp_path / "invalid_config.toml"
    toml_file.write_text(content)
    return toml_file


@pytest.fixture
def invalid_toml_file_missing(tmp_path: Path) -> Path:
    content = """
    src_dir = "data/input"
    """
    toml_file = tmp_path / "invalid_config.toml"
    toml_file.write_text(content)
    return toml_file


def test_cli_valid_toml(valid_toml_file: Path):
    Path("tests/temp_dir_cli").mkdir(exist_ok=True, parents=True)
    shutil.copy("tests/dem.tif", "tests/temp_dir_cli/dem.tif")
    result = subprocess.run(
        ["pywbt", str(valid_toml_file)],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0
    assert Path("tests", "temp_dir_cli", "dem_corr.tif").stat().st_size == 567662
    shutil.rmtree("tests/temp_dir_cli")


def test_cli_invalid_toml_missing(invalid_toml_file_missing: Path):
    result = subprocess.run(
        ["pywbt", str(invalid_toml_file_missing)],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode != 0
    assert "The TOML file must define" in result.stderr


def test_cli_invalid_toml(invalid_toml_file_not_dict: Path):
    result = subprocess.run(
        ["pywbt", str(invalid_toml_file_not_dict)],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode != 0
    assert "The TOML file must define" in result.stderr


def test_cli_missing_toml():
    result = subprocess.run(
        ["pywbt", "missing_config.toml"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode != 0
    assert "File not found" in result.stderr



"""Module to define variables and macros for the site."""

from __future__ import annotations


def define_env(env):
    """Define variables and macros for the site."""

    def issue_link(number):
        return f"[#{number}](https://github.com/cheginit/pywbt/issues/{number})"

    def commit_link(hash):
        return f"[{hash[:7]}](https://github.com/cheginit/pywbt/commit/{hash})"

    env.macro(issue_link, "issue")
    env.macro(commit_link, "commit")



"""Top-level API for PyWBT."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

from pywbt import dem_utils
from pywbt.pywbt import list_tools, prepare_wbt, tool_parameters, whitebox_tools

try:
    __version__ = version("pywbt")
except PackageNotFoundError:
    __version__ = "999"

__all__ = ["dem_utils", "list_tools", "prepare_wbt", "tool_parameters", "whitebox_tools"]



"""Configuration for pytest."""

from __future__ import annotations

import contextlib
import platform
import shutil
import tempfile

import pytest


@pytest.fixture(autouse=True)
def add_doctest_imports(doctest_namespace: dict[str, object]) -> None:
    """Add pywbt namespace for doctest."""
    import pywbt

    doctest_namespace["pywbt"] = pywbt


@pytest.fixture
def wbt_zipfile() -> str:
    """Determine the platform suffix for downloading WhiteboxTools."""
    system = platform.system()
    base_name = "WhiteboxTools_{}.zip"
    if system not in ("Windows", "Darwin", "Linux"):
        raise ValueError(f"Unsupported operating system: {system}")

    if system == "Windows":
        suffix = "win_amd64"
    elif system == "Darwin":
        suffix = "darwin_m_series" if platform.machine() == "arm64" else "darwin_amd64"
    else:
        suffix = "linux_musl" if "musl" in platform.libc_ver()[0].lower() else "linux_amd64"
    return base_name.format(suffix)


@pytest.fixture
def wrong_wbt_zipfile() -> str:
    """Determine the platform suffix for downloading WhiteboxTools."""
    system = platform.system()
    base_name = "WhiteboxTools_{}.zip"
    if system not in ("Windows", "Darwin", "Linux"):
        raise ValueError(f"Unsupported operating system: {system}")

    if system == "Windows":
        suffix = "darwin_amd64"
    elif system == "Darwin":
        suffix = "linux_amd64"
    else:
        suffix = "win_amd64"
    return base_name.format(suffix)


@pytest.fixture
def temp_dir():
    temp_dir = tempfile.mkdtemp(prefix="test_", dir=".")
    yield temp_dir
    with contextlib.suppress(PermissionError):
        shutil.rmtree(temp_dir)


@pytest.fixture
def wbt_args() -> dict[str, list[str]]:
    return {
        "BreachDepressions": ["-i=dem.tif", "--fill_pits", "-o=dem_corr.tif"],
        "D8FlowAccumulation": ["-i=dem_corr.tif", "-o=d8accum.tif"],
        "ExtractStreams": ["--flow_accum=d8accum.tif", "--threshold=600.0", "-o=streams.tif"],
    }



"""Command-line interface for pywbt."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path

from pywbt.pywbt import whitebox_tools


def main():
    parser = argparse.ArgumentParser(
        description="Run WhiteboxTools using a TOML configuration file.",
    )
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        import tomli as tomllib

    parser.add_argument("config_file", type=Path, help="Path to the TOML configuration file.")
    cfg = Path(parser.parse_args().config_file)
    if not cfg.exists():
        raise FileNotFoundError(f"File not found: {cfg}")

    try:
        with cfg.open("rb") as f:
            config = tomllib.load(f)
    except tomllib.TOMLDecodeError as e:
        raise ValueError(f"Invalid TOML file: {e}") from e

    src_dir = config.get("src_dir")
    arg_dict = config.get("arg_dict")
    files_to_save = config.get("files_to_save")
    save_dir = config.get("save_dir", ".")
    wbt_root = config.get("wbt_root", "WBT")
    compress_rasters = config.get("compress_rasters", False)
    zip_path = config.get("zip_path")
    refresh_download = config.get("refresh_download", False)
    max_procs = config.get("max_procs", -1)
    verbose = config.get("verbose", False)

    if not src_dir or not isinstance(arg_dict, dict):
        raise ValueError("The TOML file must define 'src_dir' and 'arg_dict'.")

    whitebox_tools(
        src_dir=src_dir,
        arg_dict=arg_dict,
        files_to_save=files_to_save,
        save_dir=save_dir,
        wbt_root=wbt_root,
        compress_rasters=compress_rasters,
        zip_path=zip_path,
        refresh_download=refresh_download,
        max_procs=max_procs,
        verbose=verbose,
    )



"""Utility functions for getting and processing Digital Elevation Models (DEMs)."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from pathlib import Path

    import geopandas as gpd
    import xarray as xr

    Bbox = tuple[float, float, float, float]


__all__ = ["get_3dep", "get_nasadem", "tif_to_da", "tif_to_gdf"]


class DependencyError(ImportError):
    """Raised when a required dependency is not found."""

    def __init__(self) -> None:
        self.message = "\n".join(
            (
                "The `dem_utils` module has additional dependencies that are not installed.",
                "Please install them using:",
                "`pip install 'geopandas>=1' planetary-computer pystac-client rioxarray`",
                "or:",
                "`micromamba install -c conda-forge 'geopandas-base>=1' planetary_computer pystac_client rioxarray`",
            )
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


def _check_bbox(bbox: Bbox) -> None:
    """Validate that bbox is in correct form."""
    if (
        not isinstance(bbox, Sequence)
        or len(bbox) != 4
        or not all(isinstance(x, (int, float)) for x in bbox)
    ):
        raise TypeError(
            "`bbox` must be a tuple of form (west, south, east, north) in decimal degrees."
        )


def _bbox_buffer(bbox: Bbox, buffer: float, crs_proj: int) -> tuple[Bbox, Bbox]:
    """Buffer the bounding box."""
    try:
        import shapely
        from pyproj import Transformer
    except ImportError as e:
        raise DependencyError from e
    transformer = Transformer.from_crs(4326, crs_proj, always_xy=True)
    minx, miny = transformer.transform(bbox[0], bbox[1])
    maxx, maxy = transformer.transform(bbox[2], bbox[3])
    bbox_proj = (minx, miny, maxx, maxy)

    bbox_buff = shapely.box(*bbox_proj).buffer(buffer).bounds
    transformer = Transformer.from_crs(crs_proj, 4326, always_xy=True)
    minx, miny = transformer.transform(bbox_buff[0], bbox_buff[1])
    maxx, maxy = transformer.transform(bbox_buff[2], bbox_buff[3])
    bbox_buff = (minx, miny, maxx, maxy)
    return bbox_proj, bbox_buff


def _estimate_utm(bbox: Bbox) -> int:
    """Estimate UTM CRS based on bbox."""
    try:
        from pyproj.aoi import AreaOfInterest
        from pyproj.database import query_utm_crs_info
    except ImportError as e:
        raise DependencyError from e

    minx, miny, maxx, maxy = bbox
    x_center = 0.5 * (minx + maxx)
    y_center = 0.5 * (miny + maxy)
    utm_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=x_center,
            south_lat_degree=y_center,
            east_lon_degree=x_center,
            north_lat_degree=y_center,
        ),
    )
    return int(utm_list[0].code)


def get_nasadem(bbox: Bbox, tif_path: str | Path, to_utm: bool = False) -> None:
    """Get DEM for a given bounding box from NASADEM.

    Parameters
    ----------
    bbox : tuple
        Bounding box coordinates in the order (west, south, east, north) in decimal degrees.
    tif_path : str or pathlib.Path
        Path to save the obtained data as a GeoTIFF file.
    to_utm : bool, optional
        Reproject the DEM to UTM, by default False.
    """
    try:
        import planetary_computer
        import pystac_client
        import rioxarray as rxr
        import rioxarray.merge as rxr_merge
    except ImportError as e:
        raise DependencyError from e

    _check_bbox(bbox)
    bbox_buff = bbox_utm = bbox
    utm = 4326
    if to_utm:
        utm = _estimate_utm(bbox)
        buff_size, dem_res = 20, 30
        bbox_utm, bbox_buff = _bbox_buffer(bbox, buff_size * dem_res, utm)
    catalog = pystac_client.Client.open(
        "https://planetarycomputer.microsoft.com/api/stac/v1",
        modifier=planetary_computer.sign_inplace,
    )
    signed_asset = (
        planetary_computer.sign(item.assets["elevation"]).href
        for item in catalog.search(collections=["nasadem"], bbox=bbox_buff).items()
    )
    dem = rxr_merge.merge_arrays(
        [rxr.open_rasterio(href).squeeze(drop=True) for href in signed_asset]  # pyright: ignore[reportArgumentType]
    )
    if to_utm:
        dem = dem.rio.reproject(utm).fillna(dem.rio.nodata).rio.clip_box(*bbox_utm)
    dem.attrs.update({"units": "meters", "vertical_datum": "EGM96"})
    dem.name = "elevation"
    dem.astype("int16").rio.to_raster(tif_path)
    dem.close()


def get_3dep(
    bbox: Bbox,
    tif_path: str | Path,
    resolution: int = 10,
    to_5070: bool = False,
) -> None:
    """Get DEM from USGS's 3D Hydrography Elevation Data Program (3DEP).

    Parameters
    ----------
    bbox : tuple
        Bounding box coordinates in the order (west, south, east, north)
        in decimal degrees.
    tif_path : str or pathlib.Path
        Path to save the obtained data as a GeoTIFF file.
    resolution : int, optional
        Resolution of the DEM in meters, by default 10.
    to_5070 : bool, optional
        Reproject the DEM to EPGS:5070, by default False.
    """
    try:
        import numpy as np
        import rioxarray as rxr
    except ImportError as e:
        raise DependencyError from e

    _check_bbox(bbox)
    base_url = "https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation"
    url = {
        10: f"{base_url}/13/TIFF/USGS_Seamless_DEM_13.vrt",
        30: f"{base_url}/1/TIFF/USGS_Seamless_DEM_1.vrt",
        60: f"{base_url}/2/TIFF/USGS_Seamless_DEM_2.vrt",
    }
    if resolution not in url:
        raise ValueError("Resolution must be one of 10, 30, or 60 meters.")
    bbox_buff = bbox_proj = bbox
    crs_proj = 4326
    if to_5070:
        crs_proj = 5070
        buff_size = 20
        bbox_proj, bbox_buff = _bbox_buffer(bbox, buff_size * resolution, crs_proj)

    dem = cast("xr.DataArray", rxr.open_rasterio(url[resolution]).squeeze(drop=True))
    dem = dem.rio.clip_box(*bbox_buff)
    dem = dem.where(dem > dem.rio.nodata, drop=False).rio.write_nodata(np.nan)
    if to_5070:
        dem = dem.rio.reproject(crs_proj).rio.clip_box(*bbox_proj)
    dem.attrs.update({"units": "meters", "vertical_datum": "NAVD88"})
    dem.name = "elevation"
    dem.rio.to_raster(tif_path)
    dem.close()


def tif_to_da(
    tif_path: str | Path,
    dtype: str | None = None,
    name: str | None = None,
    long_name: str | None = None,
    nodata: float | None = None,
) -> xr.DataArray:
    """Read a GeoTIFF file with a single band into an xarray DataArray.

    Parameters
    ----------
    tif_path : str or pathlib.Path
        Path to the GeoTIFF file with a single band.
    dtype : str, optional
        Data type to cast the data to, e.g. ``int32``, by default ``None``,
        i.e. use the data type from the GeoTIFF.
    name : str, optional
        Name to assign to the ``xarray.DataArray``, by default ``None``.
    long_name : str, optional
        Long name to assign to the ``xarray.DataArray``, by default ``None``.
    nodata : int or float, optional
        Value to use as ``da.rio.nodata``, by default ``None``, i.e.
        use the ``da.rio.nodata`` value from the GeoTIFF.

    Returns
    -------
    xarray.DataArray
        DataArray containing the GeoTIFF data.
    """
    try:
        import numpy as np
        import rioxarray as rxr
    except ImportError as e:
        raise DependencyError from e

    ds = rxr.open_rasterio(tif_path).squeeze(drop=True).load()
    ds.close()
    ds = cast("xr.DataArray", ds)
    if nodata is not None:
        ds = ds.rio.write_nodata(nodata)
    if dtype is None:
        dtype = str(ds.dtype)
    else:
        ds = ds.astype(dtype)
    if ds.rio.nodata is not None and np.issubdtype(dtype, np.floating):
        ds = ds.where(ds != ds.rio.nodata).rio.write_nodata(np.nan)
    if name:
        ds.attrs["name"] = name
    if long_name:
        ds.attrs["long_name"] = long_name
    return ds


def tif_to_gdf(
    tif_path: str | Path, dtype: str, feat_name: str, connectivity: int = 8
) -> gpd.GeoDataFrame:
    """Convert a GeoTIFF to a GeoDataFrame.

    Parameters
    ----------
    tif_path : str or pathlib.Path
        Path to the GeoTIFF file. The file must contain a single band.
    dtype : str
        String representation of the data type to use for the feature
        values, e.g. ``int32``.
    feat_name : str
        Name to assign to the feature values column.
    connectivity : int, optional
        Use 4 or 8 pixel connectivity for grouping pixels into features,
        by default 8.

    Returns
    -------
    geopandas.GeoDataFrame
        GeoDataFrame containing the features.
    """
    try:
        import geopandas as gpd
        import numpy as np
        import rasterio
        from rasterio import features
    except ImportError as e:
        raise DependencyError from e

    with rasterio.open(tif_path) as src:
        data = src.read(1).astype(dtype)
        nodata = np.nan if src.nodata is None else src.nodata
        feats_gen = features.shapes(
            data,
            mask=~np.isnan(data) if np.isnan(nodata) else data != nodata,
            transform=src.transform,
            connectivity=connectivity,
        )
        gdf = gpd.GeoDataFrame.from_features(
            [{"geometry": geom, "properties": {feat_name: val}} for geom, val in feats_gen],
            crs=src.crs,
        )
        gdf[feat_name] = gdf[feat_name].astype(dtype)  # pyright: ignore[reportCallIssue,reportArgumentType]
    return gdf



"""Python wrapper for WhiteboxTools command-line utility."""

from __future__ import annotations

import json
import logging
import platform
import re
import shutil
import stat
import subprocess
import urllib.error
import urllib.request
import zipfile
from functools import lru_cache
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Literal

from filelock import SoftFileLock

BASE_URL = "https://www.whiteboxgeo.com/WBT_{}/WhiteboxTools_{}.zip"

__all__ = ["list_tools", "prepare_wbt", "tool_parameters", "whitebox_tools"]

if TYPE_CHECKING:
    from subprocess import CompletedProcess

    Platform = Literal["win_amd64", "darwin_m_series", "darwin_amd64", "linux_amd64", "linux_musl"]
    System = Literal["Windows", "Darwin", "Linux"]
    ExeName = Literal["whitebox_tools.exe", "whitebox_tools"]


def _setup_logger() -> logging.Logger:
    logger = logging.getLogger("pywbt")
    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger


logger = _setup_logger()


@lru_cache(maxsize=1)
def _get_platform_suffix() -> tuple[System, Platform, ExeName]:
    """Determine the platform suffix for downloading WhiteboxTools."""
    system = platform.system()
    if system not in ("Windows", "Darwin", "Linux"):
        raise ValueError(f"Unsupported operating system: {system}")

    if system == "Windows":
        suffix = "win_amd64"
    elif system == "Darwin":
        suffix = "darwin_m_series" if platform.machine() == "arm64" else "darwin_amd64"
    else:
        suffix = "linux_musl" if "musl" in platform.libc_ver()[0].lower() else "linux_amd64"
    exe_name = "whitebox_tools.exe" if system == "Windows" else "whitebox_tools"
    return system, suffix, exe_name


def _extract_wbt(zip_path: Path, wbt_root: Path, temp_path: Path, system: System) -> None:
    """Extract WhiteboxTools from zip file."""
    try:
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(path=temp_path)

        wbt_dir = next(temp_path.glob("**/WBT"), None)
        shutil.copytree(str(wbt_dir), wbt_root, dirs_exist_ok=True)

        if system != "Windows":
            exec_path = wbt_root / "whitebox_tools"
            exec_path.chmod(exec_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        logger.info(f"Extracted WhiteboxTools to {wbt_root}")
    except zipfile.BadZipFile as e:
        raise RuntimeError("Downloaded file is not a valid zip file.") from e
    except FileNotFoundError as e:
        raise RuntimeError(f"Error extracting WhiteboxTools: {e}") from e
    except shutil.Error as e:
        raise RuntimeError(f"Error copying WhiteboxTools files: {e}") from e


@lru_cache(maxsize=1)
def _get_wbt_version(exe_path: str | Path) -> str:
    """Get the version of WhiteboxTools."""
    try:
        result = subprocess.run(
            [str(exe_path), "--version"], capture_output=True, text=True, check=True
        )
        match = re.search(r"WhiteboxTools v(\d+\.\d+\.\d+)", result.stdout.strip())
        return match.group(1) if match else "unknown"
    except (subprocess.CalledProcessError, OSError) as e:
        raise RuntimeError(f"Error running WhiteboxTools: {e!s}") from e


def _attempt_prepare_wbt(
    wbt_root: Path,
    zip_path: str | Path | None,
    refresh_download: bool,
) -> str | None:
    """Attempt to prepare WhiteboxTools once and return version or ``None`` if fails."""
    system, platform_suffix, exe_name = _get_platform_suffix()
    url = BASE_URL.format(system, platform_suffix)

    exe_path = wbt_root / exe_name
    if exe_path.exists() and not refresh_download:
        logger.info(f"Using existing WhiteboxTools executable: {exe_path}")
        try:
            return _get_wbt_version(exe_path)
        except RuntimeError as e:
            logger.warning(f"Existing executable is invalid: {e}")
            return None

    shutil.rmtree(wbt_root, ignore_errors=True)

    with TemporaryDirectory(prefix="wbt_", dir=".") as tmp:
        temp_path = Path(tmp)
        zip_path_local = temp_path / Path(url).name if zip_path is None else Path(zip_path)
        zip_path_local.parent.mkdir(parents=True, exist_ok=True)
        zip_path_local = zip_path_local.with_suffix(".zip")
        if refresh_download:
            zip_path_local.unlink(missing_ok=True)

        if not zip_path_local.exists():
            logger.info(f"Downloading WhiteboxTools from {url}")
            try:
                urllib.request.urlretrieve(url, zip_path_local)
            except urllib.error.URLError as e:
                logger.warning(f"Download failed: {e}")
                return None

        try:
            _extract_wbt(zip_path_local, wbt_root, temp_path, system)
            return _get_wbt_version(exe_path)
        except RuntimeError as e:
            logger.warning(f"Extraction failed: {e}")
            zip_path_local.unlink(missing_ok=True)
            shutil.rmtree(wbt_root, ignore_errors=True)
            return None


def prepare_wbt(
    wbt_root: str | Path,
    zip_path: str | Path | None = None,
    refresh_download: bool = False,
    max_attempts: int = 3,
) -> str:
    """Download the WhiteboxTools executable for the current platform.

    Parameters
    ----------
    wbt_root : str or Path
        Root directory to extract WhiteboxTools to.
    zip_path : str or Path, optional
        Path to the zip file containing the WhiteboxTools executables, defaults to ``None``.
        If ``None``, the zip file will be downloaded to a temporary directory that gets
        deleted after the function finishes. Otherwise, the zip file will be saved to
        the specified path.
    refresh_download : bool, optional
        Whether to refresh the download even if WhiteboxTools is found.
        Defaults to ``False``.
    max_attempts : int, optional
        Maximum number of attempts to prepare WhiteboxTools, defaults to 3.

    Returns
    -------
    str
        Version of WhiteboxTools.
    """
    wbt_root = Path(wbt_root)
    lock_path = wbt_root / ".wbt_lock"

    # Acquire lock to allow only one process to prepare WhiteboxTools
    with SoftFileLock(lock_path):
        for attempt in range(max_attempts):
            version = _attempt_prepare_wbt(wbt_root, zip_path, refresh_download)
            if version:
                return version
            logger.warning(f"Attempt {attempt + 1}/{max_attempts} failed, retrying...")

        raise RuntimeError(f"Failed to prepare WhiteboxTools after {max_attempts} attempts.")


def _run_wbt(
    exe_path: str | Path,
    tool_name: str,
    args: list[str],
    max_procs: int,
    compress_rasters: bool,
    work_dir: str | Path,
    version: str,
) -> None:
    """Run a WhiteboxTools command."""
    if tool_name in ("BreachDepressionsLeastCost", "breach_depressions_least_cost"):
        logger.warning("Forcing BreachDepressionsLeastCost to use a single process.")
        msg = " ".join(
            (
                "In WBT v2.4.0, BreachDepressionsLeastCost is unstable",
                ", for now, it is recommended to use BreachDepressions instead.",
                "For more information, see:\n",
                "https://github.com/jblindsay/whitebox-tools/issues/418\n",
                "https://github.com/jblindsay/whitebox-tools/issues/416\n",
                "https://github.com/jblindsay/whitebox-tools/issues/407\n",
            )
        )
        logger.warning(msg)
        max_procs = 1

    command = [
        str(exe_path),
        f"--run={tool_name}",
        f"--wd={work_dir}",
        f"--compress_rasters={'true' if compress_rasters else 'false'}",
        f"--max_procs={max_procs}",
        *args,
    ]

    logger.info(f"Running WhiteboxTools version: {version}")
    logger.info(f"Command: {' '.join(map(str, command))}")

    try:
        process = subprocess.run(command, check=True, text=True, capture_output=True)
        logger.info("WhiteboxTools output:")
        logger.info(process.stdout)
    except subprocess.CalledProcessError as e:
        logger.exception("WhiteboxTools error output:")
        logger.exception(e.stderr)
        logger.exception("Error running WhiteboxTools")
        raise RuntimeError(e.stderr) from e


class _WBTSession:
    """Context manager for running WhiteboxTools operations.

    Parameters
    ----------
    src_dir : Path
        Source directory containing input files.
    save_dir : Path, optional
        Directory to save the output files.
    files_to_save : tuple of str
        Files to save after the session. If ``None``, keep all intermediate files.
    """

    def __init__(
        self,
        src_dir: Path,
        save_dir: Path,
        files_to_save: list[str] | tuple[str, ...] | None,
        version: str,
    ) -> None:
        self.src_dir = src_dir
        self.files_to_save = files_to_save
        if self.files_to_save is not None:
            self.files_to_save = list(self.files_to_save)
            # add .dbf, .prj, .shx files for Shapefile outputs
            for file in self.files_to_save:
                if file.endswith(".shp"):
                    stem = file[:-4]
                    self.files_to_save.extend([f"{stem}.dbf", f"{stem}.prj", f"{stem}.shx"])

        self.save_dir = save_dir
        self.version = version
        self.outputs = {}
        self.system, *_ = _get_platform_suffix()

    @staticmethod
    def _extract_outputs(wbt_args: dict[str, list[str]]) -> set[str]:
        """Extract output filenames from WhiteboxTools arguments."""
        outputs = []
        for args in wbt_args.values():
            output = [arg.split("=")[1] for arg in args if arg.startswith(("-o=", "--output"))]
            # add .dbf, .prj, .shx files for Shapefile outputs
            for out in output:
                if out.endswith(".shp"):
                    stem = out[:-4]
                    output.extend([f"{stem}.dbf", f"{stem}.prj", f"{stem}.shx"])
            outputs.extend(output)
        return set(outputs)

    def run(
        self,
        wbt_args: dict[str, list[str]],
        wbt_root: str | Path,
        compress_rasters: bool,
        max_procs: int,
    ) -> None:
        """Run WhiteboxTools operations.

        Parameters
        ----------
        wbt_args : dict of str to list of str
            Arguments for WhiteboxTools.
        wbt_root : str or Path
            Root directory of WhiteboxTools.
        compress_rasters : bool
            Whether to compress rasters.
        max_procs : int
            Maximum number of processes to use.
        """
        _, _, exe_name = _get_platform_suffix()
        exe_path = Path(wbt_root) / exe_name
        self.outputs = self._extract_outputs(wbt_args)
        for tool_name, args in wbt_args.items():
            _run_wbt(
                exe_path, tool_name, args, max_procs, compress_rasters, self.src_dir, self.version
            )

    def _cleanup(self) -> None:
        """Delete intermediate files and move output files to save_dir."""
        if self.save_dir != self.src_dir:
            self.save_dir.mkdir(parents=True, exist_ok=True)
            if self.files_to_save is not None:
                for file in self.files_to_save:
                    source = self.src_dir / file
                    if not source.exists():
                        logger.exception(f"Output file to save {source} not found")
                        raise FileNotFoundError(f"Output file to save {source} not found")
                    destination = Path(self.save_dir, file)
                    destination.unlink(missing_ok=True)
                    shutil.move(source, destination)
                    logger.info(f"Moved output file {source} to {destination}")
                _ = [(self.src_dir / f).unlink(missing_ok=True) for f in self.outputs]
                logger.info("Deleted remaining intermediate files.")
            else:
                for file in self.outputs:
                    shutil.move(self.src_dir / file, self.save_dir)
        elif self.files_to_save is not None:
            for file in self.outputs:
                if file not in self.files_to_save:
                    Path(self.src_dir / file).unlink()
                    logger.info(f"Deleted intermediate file {file}")

    def __enter__(self) -> _WBTSession:  # noqa: PYI034
        logger.info(
            f"Starting WhiteboxTools session with source directory: {self.src_dir.absolute()}"
        )
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, *_
    ) -> bool:
        # If an exception occurred, propagate it
        if exc_type is not None:
            for file in self.outputs:
                Path(self.src_dir / file).unlink(missing_ok=True)
            logger.info("Deleted all intermediate files.")
            logger.exception(f"An error occurred: {exc_value}")
            return False
        self._cleanup()
        logger.info(
            f"Completed WhiteboxTools session with source directory: {self.src_dir.absolute()}"
        )
        return True


def whitebox_tools(
    src_dir: str | Path,
    arg_dict: dict[str, list[str]],
    files_to_save: list[str] | tuple[str, ...] | None = None,
    save_dir: str | Path = ".",
    wbt_root: str | Path = "WBT",
    compress_rasters: bool = False,
    zip_path: str | Path | None = None,
    refresh_download: bool = False,
    max_procs: int = -1,
    verbose: bool = False,
) -> None:
    """Run a WhiteboxTools (not Whitebox Runner) tool with specified arguments.

    Parameters
    ----------
    src_dir : str or Path
        Path to the source directory containing input files. This will be the working
        directory for the WhiteboxTools session. The input files should be in this directory.
        Thus, when using the files in this directory in ``arg_dict``, you must just use the
        filenames without the directory path. Refer to the example in ``arg_dict``.
    arg_dict : dict
        A dictionary containing the tool names as keys and list of each
        tool's arguments as values. For example:
        ```python
        {
            "BreachDepressions": ["-i=dem.tif", "--fill_pits", "-o=dem_corr.tif"],
            "D8Pointer": ["-i=dem_corr.tif", "-o=fdir.tif"],
            "D8FlowAccumulation": ["-i=fdir.tif", "--pntr", "-o=d8accum.tif"],
        }
        ```
        Note that the input and output file names should not contain the directory path,
        only the filenames.
    files_to_save : list, optional
        List of output files to save to ``save_dir``. Note that these should be the filenames
        without the directory path, just as they are used in the ``arg_dict``, i.e. the
        values that are passed by ``-o`` or ``--output``. Defaults to ``None``, which means
        all intermediate files will be saved to ``save_dir``.
    save_dir : str or Path, optional
        Path to the directory where the output files given in ``files_to_save`` will be saved
        (default is ``"."``, i.e., current working directory).
    wbt_root : str or Path, optional
        Path to the directory the Whitebox executables will be extracted to
        (default is ``"WBT"``).
    zip_path : str or Path, optional
        Path to the zip file containing the WhiteboxTools executables (default is ``None``).
        If ``None``, the zip file will be downloaded to a temporary directory that gets
        deleted after the function finishes. Otherwise, the zip file will be saved to
        the specified path.
    refresh_download : bool, optional
        Whether to refresh the download if WhiteboxTools is found (default is ``False``).
    compress_rasters : bool, optional
        Whether to compress output rasters (default is ``False``).
    max_procs : int, optional
        The maximum number of processes to use (default is ``-1``).
    verbose : bool, optional
        Whether to print verbose output (default is ``False``).

    Raises
    ------
    subprocess.CalledProcessError
        If the tool execution fails.
    Exception
        For any other unexpected errors.
    """
    logger.setLevel(logging.INFO if verbose else logging.WARNING)

    wbt_root = Path(wbt_root)
    save_dir = Path(save_dir) if save_dir else Path.cwd()
    save_dir.mkdir(parents=True, exist_ok=True)

    version = prepare_wbt(wbt_root, zip_path, refresh_download)

    if not isinstance(arg_dict, dict) or not all(
        isinstance(k, str) and isinstance(v, (list, tuple)) for k, v in arg_dict.items()
    ):
        raise ValueError("arg_dict must be a dict of str keys and list or tuple values.")

    if not (files_to_save is None or isinstance(files_to_save, (list, tuple))):
        raise TypeError("files_to_save must be None, a list, or tuple of strings.")

    with _WBTSession(Path(src_dir), Path(save_dir), files_to_save, version) as wbt:
        wbt.run(arg_dict, wbt_root, compress_rasters, max_procs)


def _run_wbt_cmd(
    cmd: str, wbt_root: str | Path = "WBT", zip_path: str | Path | None = None
) -> CompletedProcess[str]:
    """Run WhiteboxTools command to list available tools and their options."""
    logger.setLevel(logging.WARNING)
    _ = prepare_wbt(wbt_root, zip_path, False)
    _, _, exe_name = _get_platform_suffix()
    exe_path = Path(wbt_root) / exe_name
    return subprocess.run([str(exe_path), cmd], capture_output=True, text=True, check=True)


def list_tools(wbt_root: str | Path = "WBT", zip_path: str | Path | None = None) -> dict[str, str]:
    """List the available WhiteboxTools commands.

    Parameters
    ----------
    wbt_root : str or Path, optional
        Path to the directory containing the WhiteboxTools executables
        (default is ``"WBT"``).
    zip_path : str or Path, optional
        Path to the zip file containing the WhiteboxTools executables (default is ``None``).

    Returns
    -------
    dict
        A dictionary of WhiteboxTools commands and their descriptions. Can
        be converted to a ``pandas.Series`` for easier viewing using
        ``pd.Series(pywbt.list_tools())``.
    """
    result = _run_wbt_cmd("--listtools", wbt_root, zip_path)
    tools = (t.split(":") for t in result.stdout.strip().split("\n")[1:] if t)
    return {n: d.strip() for n, d in tools}


def tool_parameters(
    tool_name: str, wbt_root: str | Path = "WBT", zip_path: str | Path | None = None
) -> list[dict[str, str]]:
    """List the parameters for a WhiteboxTools command.

    Parameters
    ----------
    tool_name : str
        Name of the WhiteboxTools command.
    wbt_root : str or Path, optional
        Path to the directory containing the WhiteboxTools executables
        (default is ``"WBT"``).
    zip_path : str or Path, optional
        Path to the zip file containing the WhiteboxTools executables (default is ``None``).

    Returns
    -------
    list
        A list of WhiteboxTools command parameters and their descriptions.
        Can be converted to a ``pandas.DataFrame`` for easier viewing. For example,
        to get parameters of ``Aspect`` tool, use
        ``pd.DataFrame(pywbt.tool_parameters("Aspect"))``.
    """
    result = _run_wbt_cmd(f"--toolparameters={tool_name}", wbt_root, zip_path)
    return json.loads(result.stdout)["parameters"]



"""Hooks for the documentation."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from mkdocs.structure.files import File, Files

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig

changelog = Path(__file__).parent.parent / "CHANGELOG.md"
contributing = Path(__file__).parent.parent / "CONTRIBUTING.md"
readme = Path(__file__).parent.parent / "README.md"


def on_files(files: Files, config: MkDocsConfig):
    """Copy the schema to the site."""
    files.append(
        File(
            path=changelog.name,
            src_dir=changelog.parent,
            dest_dir=str(config.site_dir),
            use_directory_urls=config.use_directory_urls,
        )
    )
    files.append(
        File(
            path=contributing.name,
            src_dir=contributing.parent,
            dest_dir=str(config.site_dir),
            use_directory_urls=config.use_directory_urls,
        )
    )
    files.append(
        File(
            path=readme.name,
            src_dir=readme.parent,
            dest_dir=str(config.site_dir),
            use_directory_urls=config.use_directory_urls,
        )
    )
    return files
